#!/usr/bin/env python3

# type: ignore

"""
compare.py - versatile benchmark output compare tool
"""

import argparse
import json
import os
import sys
import unittest
from argparse import ArgumentParser

import gbench
from gbench import report, util


def check_inputs(in1, in2, flags):
    """
    Perform checking on the user provided inputs and diagnose any abnormalities
    """
    in1_kind, in1_err = util.classify_input_file(in1)
    in2_kind, in2_err = util.classify_input_file(in2)
    output_file = util.find_benchmark_flag("--benchmark_out=", flags)
    output_type = util.find_benchmark_flag("--benchmark_out_format=", flags)
    if (
        in1_kind == util.IT_Executable
        and in2_kind == util.IT_Executable
        and output_file
    ):
        print(
            (
                "WARNING: '--benchmark_out=%s' will be passed to both "
                "benchmarks causing it to be overwritten"
            )
            % output_file
        )
    if in1_kind == util.IT_JSON and in2_kind == util.IT_JSON:
        # When both sides are JSON the only supported flag is
        # --benchmark_filter=
        for flag in util.remove_benchmark_flags("--benchmark_filter=", flags):
            print(
                "WARNING: passing %s has no effect since both "
                "inputs are JSON" % flag
            )
    if output_type is not None and output_type != "json":
        print(
            (
                "ERROR: passing '--benchmark_out_format=%s' to 'compare.py`"
                " is not supported."
            )
            % output_type
        )
        sys.exit(1)


def create_parser():
    parser = ArgumentParser(
        description="versatile benchmark output compare tool"
    )

    parser.add_argument(
        "-a",
        "--display_aggregates_only",
        dest="display_aggregates_only",
        action="store_true",
        help="If there are repetitions, by default, we display everything - the"
        " actual runs, and the aggregates computed. Sometimes, it is "
        "desirable to only view the aggregates. E.g. when there are a lot "
        "of repetitions. Do note that only the display is affected. "
        "Internally, all the actual runs are still used, e.g. for U test.",
    )

    parser.add_argument(
        "--no-color",
        dest="color",
        default=True,
        action="store_false",
        help="Do not use colors in the terminal output",
    )

    parser.add_argument(
        "-d",
        "--dump_to_json",
        dest="dump_to_json",
        help="Additionally, dump benchmark comparison output to this file in JSON format.",
    )

    utest = parser.add_argument_group()
    utest.add_argument(
        "--no-utest",
        dest="utest",
        default=True,
        action="store_false",
        help="The tool can do a two-tailed Mann-Whitney U test with the null hypothesis that it is equally likely that a randomly selected value from one sample will be less than or greater than a randomly selected value from a second sample.\nWARNING: requires **LARGE** (no less than {}) number of repetitions to be meaningful!\nThe test is being done by default, if at least {} repetitions were done.\nThis option can disable the U Test.".format(
            report.UTEST_OPTIMAL_REPETITIONS, report.UTEST_MIN_REPETITIONS
        ),
    )
    alpha_default = 0.05
    utest.add_argument(
        "--alpha",
        dest="utest_alpha",
        default=alpha_default,
        type=float,
        help=(
            "significance level alpha. if the calculated p-value is below this value, then the result is said to be statistically significant and the null hypothesis is rejected.\n(default: %0.4f)"
        )
        % alpha_default,
    )

    subparsers = parser.add_subparsers(
        help="This tool has multiple modes of operation:", dest="mode"
    )

    parser_a = subparsers.add_parser(
        "benchmarks",
        help="The most simple use-case, compare all the output of these two benchmarks",
    )
    baseline = parser_a.add_argument_group("baseline", "The benchmark baseline")
    baseline.add_argument(
        "test_baseline",
        metavar="test_baseline",
        type=argparse.FileType("r"),
        nargs=1,
        help="A benchmark executable or JSON output file",
    )
    contender = parser_a.add_argument_group(
        "contender", "The benchmark that will be compared against the baseline"
    )
    contender.add_argument(
        "test_contender",
        metavar="test_contender",
        type=argparse.FileType("r"),
        nargs=1,
        help="A benchmark executable or JSON output file",
    )
    parser_a.add_argument(
        "benchmark_options",
        metavar="benchmark_options",
        nargs=argparse.REMAINDER,
        help="Arguments to pass when running benchmark executables",
    )

    parser_b = subparsers.add_parser(
        "filters", help="Compare filter one with the filter two of benchmark"
    )
    baseline = parser_b.add_argument_group("baseline", "The benchmark baseline")
    baseline.add_argument(
        "test",
        metavar="test",
        type=argparse.FileType("r"),
        nargs=1,
        help="A benchmark executable or JSON output file",
    )
    baseline.add_argument(
        "filter_baseline",
        metavar="filter_baseline",
        type=str,
        nargs=1,
        help="The first filter, that will be used as baseline",
    )
    contender = parser_b.add_argument_group(
        "contender", "The benchmark that will be compared against the baseline"
    )
    contender.add_argument(
        "filter_contender",
        metavar="filter_contender",
        type=str,
        nargs=1,
        help="The second filter, that will be compared against the baseline",
    )
    parser_b.add_argument(
        "benchmark_options",
        metavar="benchmark_options",
        nargs=argparse.REMAINDER,
        help="Arguments to pass when running benchmark executables",
    )

    parser_c = subparsers.add_parser(
        "benchmarksfiltered",
        help="Compare filter one of first benchmark with filter two of the second benchmark",
    )
    baseline = parser_c.add_argument_group("baseline", "The benchmark baseline")
    baseline.add_argument(
        "test_baseline",
        metavar="test_baseline",
        type=argparse.FileType("r"),
        nargs=1,
        help="A benchmark executable or JSON output file",
    )
    baseline.add_argument(
        "filter_baseline",
        metavar="filter_baseline",
        type=str,
        nargs=1,
        help="The first filter, that will be used as baseline",
    )
    contender = parser_c.add_argument_group(
        "contender", "The benchmark that will be compared against the baseline"
    )
    contender.add_argument(
        "test_contender",
        metavar="test_contender",
        type=argparse.FileType("r"),
        nargs=1,
        help="The second benchmark executable or JSON output file, that will be compared against the baseline",
    )
    contender.add_argument(
        "filter_contender",
        metavar="filter_contender",
        type=str,
        nargs=1,
        help="The second filter, that will be compared against the baseline",
    )
    parser_c.add_argument(
        "benchmark_options",
        metavar="benchmark_options",
        nargs=argparse.REMAINDER,
        help="Arguments to pass when running benchmark executables",
    )

    return parser


def main():
    # Parse the command line flags
    parser = create_parser()
    args, unknown_args = parser.parse_known_args()
    if args.mode is None:
        parser.print_help()
        exit(1)
    assert not unknown_args
    benchmark_options = args.benchmark_options

    if args.mode == "benchmarks":
        test_baseline = args.test_baseline[0].name
        test_contender = args.test_contender[0].name
        filter_baseline = ""
        filter_contender = ""

        # NOTE: if test_baseline == test_contender, you are analyzing the stdev

        description = "Comparing %s to %s" % (test_baseline, test_contender)
    elif args.mode == "filters":
        test_baseline = args.test[0].name
        test_contender = args.test[0].name
        filter_baseline = args.filter_baseline[0]
        filter_contender = args.filter_contender[0]

        # NOTE: if filter_baseline == filter_contender, you are analyzing the
        # stdev

        description = "Comparing %s to %s (from %s)" % (
            filter_baseline,
            filter_contender,
            args.test[0].name,
        )
    elif args.mode == "benchmarksfiltered":
        test_baseline = args.test_baseline[0].name
        test_contender = args.test_contender[0].name
        filter_baseline = args.filter_baseline[0]
        filter_contender = args.filter_contender[0]

        # NOTE: if test_baseline == test_contender and
        # filter_baseline == filter_contender, you are analyzing the stdev

        description = "Comparing %s (from %s) to %s (from %s)" % (
            filter_baseline,
            test_baseline,
            filter_contender,
            test_contender,
        )
    else:
        # should never happen
        print("Unrecognized mode of operation: '%s'" % args.mode)
        parser.print_help()
        exit(1)

    check_inputs(test_baseline, test_contender, benchmark_options)

    if args.display_aggregates_only:
        benchmark_options += ["--benchmark_display_aggregates_only=true"]

    options_baseline = []
    options_contender = []

    if filter_baseline and filter_contender:
        options_baseline = ["--benchmark_filter=%s" % filter_baseline]
        options_contender = ["--benchmark_filter=%s" % filter_contender]

    # Run the benchmarks and report the results
    json1 = json1_orig = gbench.util.sort_benchmark_results(
        gbench.util.run_or_load_benchmark(
            test_baseline, benchmark_options + options_baseline
        )
    )
    json2 = json2_orig = gbench.util.sort_benchmark_results(
        gbench.util.run_or_load_benchmark(
            test_contender, benchmark_options + options_contender
        )
    )

    # Now, filter the benchmarks so that the difference report can work
    if filter_baseline and filter_contender:
        replacement = "[%s vs. %s]" % (filter_baseline, filter_contender)
        json1 = gbench.report.filter_benchmark(
            json1_orig, filter_baseline, replacement
        )
        json2 = gbench.report.filter_benchmark(
            json2_orig, filter_contender, replacement
        )

    diff_report = gbench.report.get_difference_report(json1, json2, args.utest)
    output_lines = gbench.report.print_difference_report(
        diff_report,
        args.display_aggregates_only,
        args.utest,
        args.utest_alpha,
        args.color,
    )
    print(description)
    for ln in output_lines:
        print(ln)

    # Optionally, diff and output to JSON
    if args.dump_to_json is not None:
        with open(args.dump_to_json, "w") as f_json:
            json.dump(diff_report, f_json, indent=1)


class TestParser(unittest.TestCase):
    def setUp(self):
        self.parser = create_parser()
        testInputs = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "gbench", "Inputs"
        )
        self.testInput0 = os.path.join(testInputs, "test1_run1.json")
        self.testInput1 = os.path.join(testInputs, "test1_run2.json")

    def test_benchmarks_basic(self):
        parsed = self.parser.parse_args(
            ["benchmarks", self.testInput0, self.testInput1]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertFalse(parsed.benchmark_options)

    def test_benchmarks_basic_without_utest(self):
        parsed = self.parser.parse_args(
            ["--no-utest", "benchmarks", self.testInput0, self.testInput1]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertFalse(parsed.utest)
        self.assertEqual(parsed.utest_alpha, 0.05)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertFalse(parsed.benchmark_options)

    def test_benchmarks_basic_display_aggregates_only(self):
        parsed = self.parser.parse_args(
            ["-a", "benchmarks", self.testInput0, self.testInput1]
        )
        self.assertTrue(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertFalse(parsed.benchmark_options)

    def test_benchmarks_basic_with_utest_alpha(self):
        parsed = self.parser.parse_args(
            ["--alpha=0.314", "benchmarks", self.testInput0, self.testInput1]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.utest_alpha, 0.314)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertFalse(parsed.benchmark_options)

    def test_benchmarks_basic_without_utest_with_utest_alpha(self):
        parsed = self.parser.parse_args(
            [
                "--no-utest",
                "--alpha=0.314",
                "benchmarks",
                self.testInput0,
                self.testInput1,
            ]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertFalse(parsed.utest)
        self.assertEqual(parsed.utest_alpha, 0.314)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertFalse(parsed.benchmark_options)

    def test_benchmarks_with_remainder(self):
        parsed = self.parser.parse_args(
            ["benchmarks", self.testInput0, self.testInput1, "d"]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertEqual(parsed.benchmark_options, ["d"])

    def test_benchmarks_with_remainder_after_doubleminus(self):
        parsed = self.parser.parse_args(
            ["benchmarks", self.testInput0, self.testInput1, "--", "e"]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarks")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertEqual(parsed.benchmark_options, ["e"])

    def test_filters_basic(self):
        parsed = self.parser.parse_args(["filters", self.testInput0, "c", "d"])
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "filters")
        self.assertEqual(parsed.test[0].name, self.testInput0)
        self.assertEqual(parsed.filter_baseline[0], "c")
        self.assertEqual(parsed.filter_contender[0], "d")
        self.assertFalse(parsed.benchmark_options)

    def test_filters_with_remainder(self):
        parsed = self.parser.parse_args(
            ["filters", self.testInput0, "c", "d", "e"]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "filters")
        self.assertEqual(parsed.test[0].name, self.testInput0)
        self.assertEqual(parsed.filter_baseline[0], "c")
        self.assertEqual(parsed.filter_contender[0], "d")
        self.assertEqual(parsed.benchmark_options, ["e"])

    def test_filters_with_remainder_after_doubleminus(self):
        parsed = self.parser.parse_args(
            ["filters", self.testInput0, "c", "d", "--", "f"]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "filters")
        self.assertEqual(parsed.test[0].name, self.testInput0)
        self.assertEqual(parsed.filter_baseline[0], "c")
        self.assertEqual(parsed.filter_contender[0], "d")
        self.assertEqual(parsed.benchmark_options, ["f"])

    def test_benchmarksfiltered_basic(self):
        parsed = self.parser.parse_args(
            ["benchmarksfiltered", self.testInput0, "c", self.testInput1, "e"]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarksfiltered")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.filter_baseline[0], "c")
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertEqual(parsed.filter_contender[0], "e")
        self.assertFalse(parsed.benchmark_options)

    def test_benchmarksfiltered_with_remainder(self):
        parsed = self.parser.parse_args(
            [
                "benchmarksfiltered",
                self.testInput0,
                "c",
                self.testInput1,
                "e",
                "f",
            ]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarksfiltered")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.filter_baseline[0], "c")
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertEqual(parsed.filter_contender[0], "e")
        self.assertEqual(parsed.benchmark_options[0], "f")

    def test_benchmarksfiltered_with_remainder_after_doubleminus(self):
        parsed = self.parser.parse_args(
            [
                "benchmarksfiltered",
                self.testInput0,
                "c",
                self.testInput1,
                "e",
                "--",
                "g",
            ]
        )
        self.assertFalse(parsed.display_aggregates_only)
        self.assertTrue(parsed.utest)
        self.assertEqual(parsed.mode, "benchmarksfiltered")
        self.assertEqual(parsed.test_baseline[0].name, self.testInput0)
        self.assertEqual(parsed.filter_baseline[0], "c")
        self.assertEqual(parsed.test_contender[0].name, self.testInput1)
        self.assertEqual(parsed.filter_contender[0], "e")
        self.assertEqual(parsed.benchmark_options[0], "g")


if __name__ == "__main__":
    # unittest.main()
    main()

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# kate: tab-width: 4; replace-tabs on; indent-width 4; tab-indents: off;
# kate: indent-mode python; remove-trailing-spaces modified;



import test_make_iterator_ext as t
from common import parallelize

data = [
    {},
    { 'a' : 'b' },
    { str(i) : chr(i) for i in range(1000) }
]


def test01_key_iterator():
    for d in data:
        m = t.StringMap(d)
        assert sorted(list(m)) == sorted(list(d))


def test02_value_iterator():
    types = []
    for d in data:
        m = t.StringMap(d)
        types.append(type(m.values()))
        assert sorted(list(m.values())) == sorted(list(d.values()))
    assert types[0] is types[1] and types[1] is types[2]


def test03_items_iterator():
    for d in data:
        m = t.StringMap(d)
        assert sorted(list(m.items())) == sorted(list(d.items()))
        assert sorted(list(m.items_l())) == sorted(list(d.items()))


def test03_items_iterator_parallel(n_threads=8):
    parallelize(test03_items_iterator, n_threads=n_threads)


def test04_passthrough_iterator():
    for d in data:
        m = t.StringMap(d)
        assert list(t.iterator_passthrough(m.values())) == list(m.values())


def test05_iterator_returning_temporary():
    im = t.IdentityMap()
    assert list(im) == list(range(10))
    assert list(im.values()) == list(range(10))
    assert list(im.items()) == list(zip(range(10), range(10)))
    assert list(im.items_l()) == list(zip(range(10), range(10)))



import platform
import gc
import pytest
import threading

is_pypy = platform.python_implementation() == 'PyPy'
is_darwin = platform.system() == 'Darwin'

def collect() -> None:
    if is_pypy:
        for _ in range(3):
            gc.collect()
    else:
        gc.collect()

skip_on_pypy = pytest.mark.skipif(
    is_pypy, reason="This test currently fails/crashes PyPy")

xfail_on_pypy_darwin = pytest.mark.xfail(
    is_pypy and is_darwin, reason="This test for some reason fails on PyPy/Darwin")


# Helper function to parallelize execution of a function. We intentionally
# don't use the Python threads pools here to have threads shut down / start
# between test cases.
def parallelize(func, n_threads):
    barrier = threading.Barrier(n_threads)
    result = [None]*n_threads

    def wrapper(i):
        barrier.wait()
        result[i] = func()

    workers = []
    for i in range(n_threads):
        t = threading.Thread(target=wrapper, args=(i,))
        t.start()
        workers.append(t)

    for worker in workers:
        worker.join()
    return result




import os
import pytest
import numpy as np
import pandas as pd

from nwsspc.sharp.calc import layer
from nwsspc.sharp.calc import winds
from nwsspc.sharp.calc import thermo
from nwsspc.sharp.calc import constants


def load_parquet(filename):
    snd_df = pd.read_parquet(filename)
    snd_df = snd_df[snd_df["vwin"].notna()]
    snd_df = snd_df[snd_df["tmpc"].notna()]
    snd_df = snd_df[snd_df["relh"].notna()]
    snd_df = snd_df[snd_df["pres"] >= 50.0]

    pres = snd_df["pres"].to_numpy().astype('float32')*100.0
    hght = snd_df["hght"].to_numpy().astype('float32')
    tmpk = snd_df["tmpc"].to_numpy().astype('float32')+273.15
    dwpk = snd_df["dwpc"].to_numpy().astype('float32')+273.15
    wdir = snd_df["wdir"].to_numpy().astype('float32')
    wspd = snd_df["wspd"].to_numpy().astype('float32')
    uwin = snd_df["uwin"].to_numpy().astype('float32')
    vwin = snd_df["vwin"].to_numpy().astype('float32')

    # turn into height above ground level
    hght -= hght[0]

    # TO-DO - need a better interface to the API for doing this
    # uwin = np.empty(wspd.shape, dtype="float32")
    # vwin = np.empty(wspd.shape, dtype="float32")
    mixr = thermo.mixratio(pres, dwpk)
    vtmp = thermo.virtual_temperature(tmpk, mixr)

    return {
        "pres": pres, "hght": hght,
        "tmpk": tmpk, "mixr": mixr,
        "vtmp": vtmp, "dwpk": dwpk,
        "wdir": wdir, "wspd": wspd,
        "uwin": uwin, "vwin": vwin
    }


data_dir = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "data",
    "test_snds"
)
filename = os.path.join(
    data_dir,
    "ddc.parquet"
)

snd_data = load_parquet(filename)


def test_helicity():
    hlyr = layer.HeightLayer(0, 3000.0)
    storm_mtn = winds.WindComponents()
    storm_mtn.u = 0
    storm_mtn.v = 0

    srh = winds.helicity(
        hlyr, storm_mtn, snd_data["hght"], snd_data["uwin"], snd_data["vwin"])

    assert (srh == pytest.approx(169.71))


def test_mean_wind():

    plyr = layer.PressureLayer(snd_data["pres"][0], 60000.0)
    mw = winds.mean_wind(
        plyr, snd_data["pres"], snd_data["uwin"], snd_data["vwin"])

    assert (mw.u == pytest.approx(4.39905, abs=1e-5))
    assert (mw.v == pytest.approx(11.97378, abs=1e-5))

    mw = winds.mean_wind(
        plyr, snd_data["pres"], snd_data["uwin"], snd_data["vwin"], True)

    assert (mw.u == pytest.approx(3.683177, abs=1e-5))
    assert (mw.v == pytest.approx(11.69769, abs=1e-5))


def test_wind_shear():
    hlyr = layer.HeightLayer(0, 6000.0)
    shr = winds.wind_shear(
        hlyr, snd_data["hght"], snd_data["uwin"], snd_data["vwin"])

    assert (shr.u == pytest.approx(14.6))
    assert (shr.v == pytest.approx(13.6))

    plyr = layer.PressureLayer(snd_data["pres"][0], 50000.0)
    shr = winds.wind_shear(
        plyr, snd_data["pres"], snd_data["uwin"], snd_data["vwin"])

    assert (shr.u == pytest.approx(19.810976))
    assert (shr.v == pytest.approx(8.3890247))


def test_vector():
    wcomp = winds.WindComponents(12.0, 0.0)

    wspd = winds.vector_magnitude(constants.MISSING, wcomp.v)
    wdir = winds.vector_angle(constants.MISSING, wcomp.v)
    assert (wspd == constants.MISSING)
    assert (wdir == constants.MISSING)

    wspd = winds.vector_magnitude(wcomp.u, constants.MISSING)
    wdir = winds.vector_angle(wcomp.u, constants.MISSING)
    assert (wspd == constants.MISSING)
    assert (wdir == constants.MISSING)

    wspd = winds.vector_magnitude(wcomp.u, wcomp.v)
    wdir = winds.vector_angle(wcomp.u, wcomp.v)

    assert (wspd == wcomp.u)
    assert (wdir == 270)

    wvec = winds.components_to_vector(wcomp.u, wcomp.v)

    assert (wvec.speed == wcomp.u)
    assert (wvec.direction == 270)

    wvec = winds.components_to_vector(wcomp)

    assert (wvec.speed == wcomp.u)
    assert (wvec.direction == 270)

    wspd, wdir = winds.components_to_vector(snd_data["uwin"], snd_data["vwin"])

    assert (wspd == pytest.approx(snd_data["wspd"]))
    assert (wdir == pytest.approx(snd_data["wdir"]))


def test_components():
    wvec = winds.WindVector(60.0, 270.0)

    u_comp = winds.u_component(constants.MISSING, wvec.direction)
    v_comp = winds.v_component(constants.MISSING, wvec.direction)
    assert (u_comp == constants.MISSING)
    assert (v_comp == constants.MISSING)

    u_comp = winds.u_component(wvec.speed, constants.MISSING)
    v_comp = winds.v_component(wvec.speed, constants.MISSING)
    assert (u_comp == constants.MISSING)
    assert (v_comp == constants.MISSING)

    u_comp = winds.u_component(wvec.speed, wvec.direction)
    v_comp = winds.v_component(wvec.speed, wvec.direction)

    assert (u_comp == 60.0)
    assert (v_comp == pytest.approx(0.0, abs=1e-6))

    wcmp = winds.vector_to_components(wvec)

    assert (wcmp.u == 60.0)
    assert (wcmp.v == pytest.approx(0.0, abs=1e-6))

    wcmp = winds.vector_to_components(wvec.speed, wvec.direction)

    assert (wcmp.u == 60.0)
    assert (wcmp.v == pytest.approx(0.0, abs=1e-6))

    uwin, vwin = winds.vector_to_components(snd_data["wspd"], snd_data["wdir"])

    assert (uwin == pytest.approx(snd_data["uwin"], abs=2e-5))
    assert (vwin == pytest.approx(snd_data["vwin"], abs=2e-5))



import sys
import os

if sys.version_info < (3, 8):
    raise ImportError("nanobind does not support Python < 3.8.")

def source_dir() -> str:
    "Return the path to the nanobind source directory."
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "src")

def include_dir() -> str:
    "Return the path to the nanobind include directory"
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "include")

def cmake_dir() -> str:
    "Return the path to the nanobind CMake module directory."
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "cmake")

__version__ = "2.7.1-dev1"

__all__ = (
    "__version__",
    "source_dir",
    "include_dir",
    "cmake_dir",
)



import test_thread_ext as t
from test_thread_ext import Counter, GlobalData, ClassWithProperty, ClassWithClassProperty
from common import parallelize

def test01_object_creation(n_threads=8):
    # This test hammers 'inst_c2p' from multiple threads, and
    # checks that the locking of internal data structures works

    n = 100000
    def f():
        r = [None]*n
        for i in range(n):
            c = Counter()
            c.inc_unsafe()
            r[i] = c
        for i in range(n):
            assert t.return_self(r[i]) is r[i]
        return r

    v = parallelize(f, n_threads=n_threads)
    assert len(v) == n_threads
    for v2 in v:
        assert len(v2) == n
        for v3 in v2:
            assert v3.value == 1

def test02_global_lock(n_threads=8):
    # Test that a global PyMutex protects the counter
    n = 100000
    c = Counter()
    def f():
        for _ in range(n):
            t.inc_global(c)

    parallelize(f, n_threads=n_threads)
    assert c.value == n * n_threads


def test03_locked_method(n_threads=8):
    # Checks that nb::lock_self() protects an internal counter
    n = 100000
    c = Counter()
    def f():
        for i in range(n):
            c.inc_safe()

    parallelize(f, n_threads=n_threads)
    assert c.value == n * n_threads


def test04_locked_function(n_threads=8):
    # Checks that nb::lock_self() protects an internal counter
    n = 100000
    c = Counter()
    def f():
        for _ in range(n):
            t.inc_safe(c)

    parallelize(f, n_threads=n_threads)
    assert c.value == n * n_threads


def test05_locked_twoargs(n_threads=8):
    # Check two-argument locking
    n = 100000
    c = Counter()
    def f():
        c2 = Counter()
        for i in range(n):
            c2.inc_unsafe()
            if i & 1 == 0:
                c2.merge_safe(c)
            else:
                c.merge_safe(c2)

    parallelize(f, n_threads=n_threads)
    assert c.value == n * n_threads


def test06_global_wrapper(n_threads=8):
    # Check wrapper lookup racing with wrapper deallocation
    n = 10000
    def f():
        for _ in range(n):
            GlobalData.get()
            GlobalData.get()
            GlobalData.get()
            GlobalData.get()

    parallelize(f, n_threads=n_threads)


def test07_access_attributes(n_threads=8):
    n = 1000
    c1 = ClassWithProperty(123)
    c2 = ClassWithClassProperty(c1)

    def f():
        for i in range(n):
            _ = c2.prop1.prop2

    parallelize(f, n_threads=n_threads)



import test_exception_ext as t
import pytest

def test01_base():
    with pytest.raises(RuntimeError):
        assert t.raise_generic()

def test02_bad_alloc():
    with pytest.raises(MemoryError):
        assert t.raise_bad_alloc()

def test03_runtime_error():
    with pytest.raises(RuntimeError) as excinfo:
        assert t.raise_runtime_error()
    assert str(excinfo.value) == 'a runtime error'

def test04_domain_error():
    with pytest.raises(ValueError) as excinfo:
        assert t.raise_domain_error()
    assert str(excinfo.value) == 'a domain error'

def test05_invalid_argument():
    with pytest.raises(ValueError) as excinfo:
        assert t.raise_invalid_argument()
    assert str(excinfo.value) == 'an invalid argument error'

def test06_length():
    with pytest.raises(ValueError) as excinfo:
        assert t.raise_length_error()
    assert str(excinfo.value) == 'a length error'

def test07_out_of_range():
    with pytest.raises(IndexError) as excinfo:
        assert t.raise_out_of_range()
    assert str(excinfo.value) == 'an out of range error'

def test08_range_error():
    with pytest.raises(ValueError) as excinfo:
        assert t.raise_range_error()
    assert str(excinfo.value) == 'a range error'

def test09_overflow_error():
    with pytest.raises(OverflowError) as excinfo:
        assert t.raise_overflow_error()
    assert str(excinfo.value) == 'an overflow error'

def test10_index_error():
    with pytest.raises(IndexError) as excinfo:
        assert t.raise_index_error()
    assert str(excinfo.value) == 'an index error'

def test11_key_error():
    with pytest.raises(KeyError) as excinfo:
        assert t.raise_key_error()
    assert str(excinfo.value) == "'a key error'"

def test12_value_error():
    with pytest.raises(ValueError) as excinfo:
        assert t.raise_value_error()
    assert str(excinfo.value) == 'a value error'

def test13_type_error():
    with pytest.raises(TypeError) as excinfo:
        assert t.raise_type_error()
    assert str(excinfo.value) == 'a type error'

def test14_import_error():
    with pytest.raises(ImportError) as excinfo:
        assert t.raise_import_error()
    assert str(excinfo.value) == 'an import error'

def test15_attribute_error():
    with pytest.raises(AttributeError) as excinfo:
        assert t.raise_attribute_error()
    assert str(excinfo.value) == 'an attribute error'

def test16_stop_iteration():
    with pytest.raises(StopIteration) as excinfo:
        assert t.raise_stop_iteration()
    assert str(excinfo.value) == 'a stop iteration error'

def test17_raise_my_error_1():
    with pytest.raises(RuntimeError) as excinfo:
        assert t.raise_my_error_1()
    assert str(excinfo.value) == 'MyError1'

def test18_raise_my_error_2():
    with pytest.raises(IndexError) as excinfo:
        assert t.raise_my_error_2()
    assert str(excinfo.value) == 'MyError2'

def test19_raise_my_error_3():
    with pytest.raises(t.MyError3) as excinfo:
        assert t.raise_my_error_3()
    assert str(excinfo.value) == 'MyError3'

def test20_nested():
    def foo(arg):
        return arg / 0
    with pytest.raises(RuntimeError) as excinfo:
        t.raise_nested(foo)
    assert str(excinfo.value) == 'Call with value 123 failed'
    assert str(excinfo.value.__cause__) == 'division by zero'

